import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.image as mpimg
from matplotlib import ticker
import pickle
import json
import os

# preferred .png format for more information
# https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html
# https://matplotlib.org/stable/tutorials/introductory/images.html?highlight=mpimg
image_extension_list = ['.bmp', '.dib', '.eps', '.gif', '.icns', '.ico', '.im', '.jpeg', '.msp', '.pcx', '.png', 'ppm',
                   '.sgi', '.spider', '.tga', 'tiff', 'webp', '.xbm']


# code adapt from stackoverflow https://stackoverflow.com/questions/42481203/heatmap-on-top-of-image
def _transparent_cmap(cmap, n=255):
    """transparent heat map color map"""
    mycmap = cmap
    mycmap._init()
    mycmap._lut[:, -1] = np.linspace(0, 0.8, n + 4)
    return mycmap


def _gaussian2d(coordinates, center, sigma_x, sigma_y):
    """simple Gaussian function for visualization"""
    x, y = coordinates
    xo, yo = center
    a = 1./(2*sigma_x**2) + 1./(2*sigma_y**2)
    c = 1./(2*sigma_x**2) + 1./(2*sigma_y**2)
    g = np.exp( - (a*((x-xo)**2) + c*((y-yo)**2)))
    return g


def _load_keyboard_map(key_plot_map_directory):
    """keyboard style parser"""
    # directory input correction
    key_plot_map_file = None
    img_file = None
    if key_plot_map_directory[-1] != '/':
        key_plot_map_directory = key_plot_map_directory + '/'
    file_list = os.listdir(key_plot_map_directory)
    for file_name in file_list:
        if ".json" in file_name:
            key_plot_map_file = file_name
        elif os.path.splitext(file_name)[1].lower() in image_extension_list:
            img_file = file_name
    if key_plot_map_file and img_file:
        with open(key_plot_map_directory + key_plot_map_file, 'r') as f:
            key_plot_map = json.load(f)

        coor_key_map = {}
        coordinate = key_plot_map['coordinate']
        scan_code = key_plot_map['scan_code']
        for k in coordinate:
            row = coordinate[k]
            y = row[0]
            for i in range(len(row[1])):
                x = row[1][i]
                coor_key_map[scan_code[k][i]] = (y, x)

        kb_img = mpimg.imread(key_plot_map_directory + img_file)
        return coor_key_map, kb_img
    else:
        print("file(s) missing")


def visualize_key_stroke(records, keyboard_style=None, keyboard_offset=(0, 0), exclude_key_list=[],
                         axis_handle=plt, numeric='freq', log_scale=False):
    """
    arguments
    records (dict-like):    is dictionary generated by RecordingSession
                            (flatten current version of records is not supported)
    keyboard_style (str):   specified plotted keyboard image (default: None)
                                options
                                'MainType/blank'
                                'MainType/Thai_Pattachoat_shift'
                                'MainType/Thai_Pattachoat_no_shift'
    keyboard_offset (tuple(int, int)):  specified offset when keyboard_style key coordinates
                                        are not align with keyboard image, it design as a quick solution
                                        but directly editing the associated key_plot_map.json file is preferred
                                        (default: (0, 0))
    exclude_key_list (list): list of scan_code which is not desired to be displayed (default: empty list)
    axis_handle (matplotlib.pyplot.Axes): 2d Axes which the visualize are drawn (default: matplotlib.pyplot)
    numeric (str): specifies numeric values are displayed on the plot. (default 'freq')
                options
                'prop': proportion
                'percent': percent
                'freq': frequency
    log_scale (bool):   specify weather to log the key stroke frequencies. being used when
                        the frequency of keys are more than 10 times different, the linear scale will
                        only shown the most frequently pressed keys to lower the sensitivity of linear scale.
                        (default: False)
    """
    # load key_map associated with keyboard_styles pictures
    if not keyboard_style:
        keyboard_style = os.path.dirname(__file__) + '/keyboard_styles/MainType/blank'
    key_map, kb_img = _load_keyboard_map(keyboard_style)

    # process record
    total = 0
    for k in exclude_key_list:
        key_map.pop(k)
    for k in key_map:
        if k not in records:
            key_map[k] = (key_map[k], 0)
        else:
            value = records[k]
            key_map[k] = (key_map[k], value)
            total += value

    h, w, d = kb_img.shape
    y, x = np.mgrid[0:h, 0:w]

    xy_z = list(key_map.values())
    zz = np.zeros_like(x)
    for m in xy_z:
        high = m[1]
        if high != 0:
            if log_scale:
                high = np.log10(high)
            x_offset, y_offset = keyboard_offset
            cy, cx = m[0]
            cx += x_offset
            cy += y_offset
            zz = zz + high * _gaussian2d((x, y), (cx, cy), .07 * x.max(), .07 * y.max())

    # Plot image and overlay colormap
    mycmap = _transparent_cmap(plt.cm.jet)
    axis_handle.imshow(kb_img)
    axis_handle.contourf(x, y, zz, 15, cmap=mycmap)
    # remove ticks
    axis_handle.set_xticks([])
    axis_handle.set_yticks([])
    # axis_handle.xticks([], [])
    # axis_handle.yticks([], [])
    # preparing label data
    value_label_dict = {'freq': [], 'prop': [], 'percent': []}
    for m in xy_z:
        xy = m[0]
        # frequency
        value_label_dict['freq'].append([xy, "%d" % m[1]])
        # proportion
        value_label_dict['prop'].append([xy, "%.3f" % (m[1]/total)])
        # percent
        value_label_dict['percent'].append([xy, "%.2f%%" % (m[1]*100/total)])
    for m in value_label_dict[numeric]:
        y, x = m[0]
        axis_handle.text(x, y, m[1], ha='center')
    return 0


def visualize_finger_load(records, axis_handle=plt, numeric='freq', exclude_shift=True):
    """
    visualize finger load of key strokes records by plotting

    arguments
    records (dict-like): key stroke records generated by kbana.capture.RecordSession or kbana.analysis.simulate_record
    axis_handle (matplotlib.pyplot.Axes):  2d Axes which the visualize are drawn (default: matplotlib.pyplot)
    numeric (str): specifies numeric values are displayed on the plot. (default 'freq')
                options
                'prop': proportion
                'percent': percent
                'freq': frequency
    exclude_shift (bool): specifies if shift key pressed are taken into account of the plot
    """
    finger_scode_map_json = os.path.dirname(__file__) + '/maps/finger_scode_map.json'
    with open(finger_scode_map_json, 'r') as f:
        finger_scode_map = json.load(f)
    # convert finger_scode_map key to int
    finger_scode_map_temp = {}
    for k in finger_scode_map:
        finger_scode_map_temp[int(k)] = finger_scode_map[k]
    finger_scode_map = finger_scode_map_temp
    if exclude_shift:
        # remove left shift
        finger_scode_map.pop(47)
        # remove right shift
        finger_scode_map.pop(54)

    # prepare finger_frequency
    total = 0
    finger_set = set(finger_scode_map.values())
    finger_frequency = {}
    for finger in finger_set:
        finger_frequency[finger] = 0
    for k in records:
        # exclude non character keys
        if k in finger_scode_map:
            finger = finger_scode_map[k]
            # calculate finger analysis
            value = records[k]
            finger_frequency[finger] += value
            total += value

    if total == 0:
        raise Exception("The record is empty")

    finger_image_path = os.path.dirname(__file__) + '/misc/hands.png'
    hands = mpimg.imread(finger_image_path)
    h, w, d = hands.shape
    # create grid array
    y, x = np.mgrid[0:h, 0:w]
    z = np.zeros_like(x)

    finger_points = {'l_pinky': (145, 384),
                     'r_pinky': (1264, 384),
                     'l_ring': (225, 211),
                     'r_ring': (1179, 211),
                     'l_middle': (350, 154),
                     'r_middle': (1059, 154),
                     'l_index': (498, 199),
                     'r_index': (905, 199)}

    # color map
    mycmap = _transparent_cmap(plt.cm.jet)
    # plot hands
    # hand load
    left = 0
    right = 0
    axis_handle.imshow(hands)
    # value_label_dict = {'freq': [], 'prop': [], 'percent': []}
    for finger in finger_frequency:
        high = finger_frequency[finger]
        finger_coordinate = finger_points[finger]
        if numeric == "prop":
            axis_handle.text(*finger_coordinate, "%.3f" % (high/total))
        elif numeric == 'percent':
            axis_handle.text(*finger_coordinate, "%.2f%%" % (high*100 / total))
        elif numeric == 'freq':
            axis_handle.text(*finger_coordinate, "%d" % high)
        if finger[0] == 'l':
            left += high
        elif finger[0] == 'r':
            right += high
        z = z + high * _gaussian2d((x, y), finger_points[finger], .04 * x.max(), .04 * y.max())
    # plot hand loads
    left_xy = (317, 655)
    right_xy = (1083, 655)
    value_hand_label_dict = {
        'prop': [
            [left_xy, "%.3f" % (left/total)], [right_xy, "%.3f" % (right/total)]
        ], 'percent': [
            [left_xy, "%.2f%%" % (left * 100 / total)], [right_xy, "%.2f%%" % (right * 100 / total)]
        ], 'freq': [
            [left_xy, "%d" % (left)], [right_xy, "%d" % (right)]
        ]}
    for m in value_hand_label_dict[numeric]:
        axis_handle.text(*m[0], m[1], ha='center')

    axis_handle.contourf(x, y, z, 15, cmap=mycmap)
    # remove ticks
    axis_handle.set_xticks([])
    axis_handle.set_yticks([])
    # axis_handle.xticks([], [])
    # axis_handle.yticks([], [])
    return finger_frequency


def simulate_records(text, layout):
    """
    simulate keystroke of input text according to keyboard layout using scan code map
    Note : space and enter are ignored

    arguments
    text (str): a string of characters
    layout  (str): specify layout to simulate input data currently support layout options are
                Thai: 'kedmanee', 'pattachoat'
                Endlish: 'qwerty'
    """
    if layout is None:
        raise ValueError("layout must be stringType")
    scode_name_map_path = os.path.dirname(__file__) + '/maps/scode_name_map.json'
    with open(scode_name_map_path, 'r') as f:
        scode_name_map = json.load(f)

    shift_key_name_map_path = os.path.dirname(__file__) + '/maps/shift_key_name_map.json'
    with open(shift_key_name_map_path, 'r') as f:
        shift_key_name_map = json.load(f)

    scode_name_map = scode_name_map[layout]
    # enable simulate shift key pressed in available layout
    if layout in shift_key_name_map:
        shift_key_name_map = shift_key_name_map[layout]
    else:
        shift_key_name_map = None
    records = {}
    c_previous = None
    for c in text:
        # if simulate shift key is available then simulate shift key pressed
        if shift_key_name_map is not None:
            # verify if the character is shifted character
            if c in shift_key_name_map:
                # if previous character are shifted character
                # then assume the shifted is hold and not add shift key to the simulated records
                if not c_previous or c_previous not in shift_key_name_map:
                    shift_key = shift_key_name_map[c]
                    if shift_key in records:
                        records[shift_key] += 1
                    else:
                        records[shift_key] = 1
        # map character into scan code
        if c in scode_name_map:
            key_stroke = scode_name_map[c]
            if key_stroke in records:
                records[key_stroke] += 1
            else:
                records[key_stroke] = 1

        c_previous = c
        if len(records) == 0:
            raise Exception('none of input text characters are recognized in this layout')
    return records


def load_words_from_file(path_to_file, allow_duplicate=False):
    """
    load words which are contained in text file and removes duplicates only works if each words are separated.

    arguments
    path_to_file (str): absolute path to text file
    allow_duplicate (bool): leaves duplicated words as is (default False) set as True when loading articles
    """
    with open(path_to_file, 'r', encoding='utf-8') as f:
        words = f.read()
    words = words.split(' ')
    if not allow_duplicate:
        # because set cannot have duplicate values
        words = list(set(words))
    words = ' '.join(words)
    return words


class Analysis:
    """
    class for gui development
    """
    def __init__(self, keyboard_style=None):
        if not keyboard_style:
            self.keyboard_style = os.path.dirname(__file__) + '/keyboard_styles/MainType/blank'

    def visualize_key_stroke(self, records, axis_handle=plt, *args, **kwargs):
        return visualize_key_stroke(records, self.keyboard_style, axis_handle=axis_handle, *args, **kwargs)




import matplotlib.pyplot as plt
from .analysis import visualize_key_stroke, visualize_finger_load, simulate_records


def quick_plot(text, layout=None, numeric='freq'):
    """
    text (str, dict):
    layout (str):
    support layouts
        Thai: kebmanee, pattachoat
        English: qwerty

    please specify layout so that input text can be mapped to scan codes associated to the layout
    """
    if type(text) == dict:
        records = text
    elif type(text) == str:
        records = simulate_records(text, layout)
    else:
        raise ValueError('text must be stringType or dictionaryType')
    fig, ax = plt.subplots(2)
    visualize_key_stroke(records, log_scale=False, axis_handle=ax[0],
                         exclude_key_list=[29, 91, 56, 57, 14, 42, 54, 15, 58, 28, 'n/a'],
                         numeric=numeric)
    visualize_finger_load(records, axis_handle=ax[1], numeric=numeric)
    return fig



import keyboard
import json
import pickle
import os
import datetime
import copy
import ctypes


def get_keyboard_language():
    # only language not variant layout
    user32 = ctypes.WinDLL('user32', use_last_error=True)
    curr_window = user32.GetForegroundWindow()
    thread_id = user32.GetWindowThreadProcessId(curr_window, 0)
    klid = user32.GetKeyboardLayout(thread_id)
    lid = klid & (2 ** 16 - 1)
    lid_hex = f"{lid:#0{6}x}"
    root = os.path.dirname(__file__)
    with open(root + '/misc/win-language-id.json', 'r') as f:
        win_layout = json.load(f)
    return win_layout[lid_hex]


def load_records(filename):
    with open(filename, "rb") as f:
        records = pickle.load(f)
    return records


def flatten_records(records):
    records = records
    combined = records[False]
    shifted = records[True]
    for k in shifted:
        if k in combined:
            combined[k] += shifted[k]
        else:
            combined[k] = shifted[k]
    return combined


class RecordingSession:
    def __init__(self, filename=None, directory=None):
        self.time_stamp = datetime.datetime.now().strftime("%y_%m_%d-%H_%M_%S")
        self._shift_state = False
        if filename:
            # if filename is provided ignore directory
            self._filename = filename
        elif directory:
            if not os.path.exists(directory):
                os.mkdir(directory)
            if directory[-1] != '/':
                directory = directory + '/'
            self._directory = directory
            self._filename = directory + filename
        else:
            # if both are not provided generate filenames and records directory
            # relative to the file that call this module
            if not os.path.exists('./records'):
                os.mkdir('./records')
            self._filename = "./records/recordings-" + self.time_stamp + '.pyd'
        # if provided filename exists continue record from the recording otherwise create new records
        if os.path.exists(self._filename):
            records = load_records(self._filename)

        else:
            records = {}
        self._records = records
        # list of key strokes
        self._records_2 = []

    def _record_key_preserve_order(self, key_event):
        record_book = self._records_2
        if key_event.event_type == 'up':
            k = key_event.scan_code
            record_book.append(k)

    def _shift_toggle_preserve_order(self, key_event):
        shift_key = key_event.scan_code
        if key_event.event_type == 'down':
            self._records_2.append(shift_key)

    def _record_key_just_count(self, key_event):
        record_book = self._records
        if key_event.event_type == 'up':
            k = key_event.scan_code
            if k in record_book:
                record_book[k] += 1
            else:
                record_book[k] = 1

    def _shift_toggle_just_count(self, key_event):
        shift_key = key_event.scan_code
        if key_event.event_type == 'down':
            if shift_key in self.records:
                self.records[shift_key] += 1
            else:
                self.records[shift_key] = 1

    def _record_key(self, key_event, mode='just_count'):
        if mode == 'just_count':
            self._record_key_just_count(key_event)
        elif mode == 'preserve_order':
            self._record_key_preserve_order(key_event)

    def _shift_toggle(self, key_event, mode='just_count'):
        if mode == 'just_count':
            self._shift_toggle_just_count(key_event)
        if mode == 'preserve_order':
            self._shift_toggle_preserve_order(key_event)

    def record_built_in(self):
        while True:
            x = keyboard.read_event()
            # shift key hold
            if 'shift' in x.name:
                self._shift_toggle(x)
                while True:
                    x = keyboard.read_event()
                    if ('shift' in x.name) and (x.event_type == 'up'):
                        self._shift_toggle(x)
                        break
                    elif not ('shift' in x.name and x.event_type == 'down'):
                        # print(f"{x.scan_code} {x.name} {x.event_type} ")
                        self._record_key(x)
            # control key hold
            elif 'ctrl' in x.name:
                while True:
                    x = keyboard.read_event()
                    if ('ctrl' in x.name) and (x.event_type == 'up'):
                        break
                    elif not ('ctrl' in x.name and x.event_type == 'down'):
                        # print(f"{x.scan_code} {x.name} {x.event_type} ")
                        self._record_key(x)
            # alternate key hold
            elif 'alt' in x.name:
                while True:
                    x = keyboard.read_event()
                    if ('alt' in x.name) and (x.event_type == 'up'):
                        break
                    elif ('shift' in x.name) and (x.event_type == 'up'):
                        # print("change language")
                        language = get_keyboard_language()
                    elif not ('alt' in x.name and x.event_type == 'down'):
                        # print(f"{x.scan_code} {x.name} {x.event_type} ")
                        self._record_key(x)
            # windows key hold
            elif 'windows' in x.name:
                while True:
                    x = keyboard.read_event()
                    if ('windows' in x.name) and (x.event_type == 'up'):
                        break
                    elif ('space' in x.name) and (x.event_type == 'up'):
                        # print("change language")
                        language = get_keyboard_language()
                    elif not ('windows' in x.name and x.event_type == 'down'):
                        # print(f"{x.scan_code} {x.name} {x.event_type} ")
                        self._record_key(x)
            # ordinary key stroke
            # print(f"{x.scan_code} {x.name} {x.event_type} ")
            self._record_key(x)
            if keyboard.is_pressed("esc"):
                break

    def record(self):
        x = keyboard.read_event()
        # shift key hold
        if 'shift' in x.name:
            self._shift_toggle(x)
            while True:
                x = keyboard.read_event()
                if ('shift' in x.name) and (x.event_type == 'up'):
                    self._shift_toggle(x)
                    break
                elif not ('shift' in x.name and x.event_type == 'down'):
                    # print(f"{x.scan_code} {x.name} {x.event_type} ")
                    self._record_key(x)
        # control key hold
        elif 'ctrl' in x.name:
            while True:
                x = keyboard.read_event()
                if ('ctrl' in x.name) and (x.event_type == 'up'):
                    break
                elif not ('ctrl' in x.name and x.event_type == 'down'):
                    # print(f"{x.scan_code} {x.name} {x.event_type} ")
                    self._record_key(x)
        # alternate key hold
        elif 'alt' in x.name:
            while True:
                x = keyboard.read_event()
                if ('alt' in x.name) and (x.event_type == 'up'):
                    break
                elif ('shift' in x.name) and (x.event_type == 'up'):
                    # print("change language")
                    language = get_keyboard_language()
                elif not ('alt' in x.name and x.event_type == 'down'):
                    # print(f"{x.scan_code} {x.name} {x.event_type} ")
                    self._record_key(x)
        # windows key hold
        elif 'windows' in x.name:
            while True:
                x = keyboard.read_event()
                if ('windows' in x.name) and (x.event_type == 'up'):
                    break
                elif ('space' in x.name) and (x.event_type == 'up'):
                    # print("change language")
                    language = get_keyboard_language()
                elif not ('windows' in x.name and x.event_type == 'down'):
                    # print(f"{x.scan_code} {x.name} {x.event_type} ")
                    self._record_key(x)
        # ordinary key stroke
        # print(f"{x.scan_code} {x.name} {x.event_type} ")
        self._record_key(x)
        return 0

    def save_recording(self, filename=None):
        if filename:
            with open(filename, "wb") as f:
                pickle.dump(self._records, f)
        else:
            with open(self._filename, "wb") as f:
                pickle.dump(self._records, f)

    # def records_key_only(self):
    #     records = self.records
    #     combined = records[False]
    #     shifted = records[True]
    #     for k in shifted:
    #         if k in combined:
    #             combined[k] += shifted[k]
    #         else:
    #             combined[k] = shifted[k]
    #     return combined

    # use property decorator so that user cannot directly modified records
    @property
    def records(self):
        return copy.deepcopy(self._records)

    @property
    def filename(self):
        return self._filename



import setuptools
setuptools.setup(
    include_package_data=True,
    packages=['kbana'],
    pakage_dir={"": "src"},
    # using MANIFEST.in instead
    # package_data={
    #     "kbana": ["analysis/maps/*.json",
    #               "analysis/misc/*.png",
    #               "analysis/keyboard_styles/MainType/blank/*",
    #               "capture/misc/*.json"]
    # }
)
